/*****************************************************************************\
 *                                                                           *
 * Generator.h                                                               *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
 * Defines the class which generates xrttigen output given Parser input      *
 *                                                                           *
\*****************************************************************************/

#ifndef GENERATOR_H
#define GENERATOR_H

#include "Configuration.h"
#include "Parser.h"

#include "Essential/Log.h"

using namespace Essential;

class Generator
{
public:

    // Creates a Generator which will generate from the content in a Parser
    Generator(const Configuration &config, const Parser &parser);

    // Reads the data in from the Parser and generates output according to
    // the given Configuration
    void Generate();

private:

    class Class;
    class Field;
    class Method;
    
    // -----------------------------------------------------------------------
    // Base class of all Generator objects, all of which are based on
    // a Parser::Element
    // -----------------------------------------------------------------------
    class Base : public RefCounted
    {
    protected:

        Base(class Generator &generator, const String &id)
            : generatorM(generator), 
              pElementM(generator.parserM.LookupElement(id))
        {
            if (pElementM == NULL) {
                generator.Error("Missing id: %S\n", (const unichar *) id);
            }
        }

        Base(class Generator &generator,
             constref<Parser::Element> pElement)
            : generatorM(generator), pElementM(pElement)
        {
        }

        class Generator &generatorM;
        constref<Parser::Element> pElementM;
    };

    class Registered : public Base
    {
    public:

        Registered(class Generator &generator, const String &id)
            : Base(generator, id), hasBeenEmittedM(false) { }

        enum RegisteredType
        {
            RegisteredTypeContext,
            RegisteredTypeEnumeration,
            RegisteredTypeType
        };

        void SetNumber(U32 number)
        {
            numberM = number;
        }

        U32 GetNumber()
        {
            return numberM;
        }
        
        // Only returns false once, and then returns true thereafter
        bool ShouldEmit()
        {
            if (hasBeenEmittedM) {
                return false;
            }

            hasBeenEmittedM = true;

            return true;
        }

        bool HasBeenEmitted()
        {
            return hasBeenEmittedM;
        }

    private:

        U32 numberM;
        bool hasBeenEmittedM;
    };

    // -----------------------------------------------------------------------
    // Context is the context in which a data structure is defined;
    // it has a parent context too (unless it is the root Context (::), in
    // which case it has no parent and GetContext() returns NULL).
    // -----------------------------------------------------------------------
    class Context : public Registered
    {
    public:

        enum ContextType
        {
            ContextTypeClass,
            ContextTypeNamespace
        };

        virtual RegisteredType GetRegisteredType()
        {
            return RegisteredTypeContext;
        }

        virtual ContextType GetContextType() = 0;

        virtual void Emit() = 0;

        String GetName()
        {
            // gccxml apparently has decided to forego names on some
            // contexts now.  The name can be recovered from the "demangled"
            // field though
            String name = pElementM->GetAttributeValue("name");
            if (name == "") {
                String demangled = pElementM->GetAttributeValue("demangled");
                if (demangled == "") {
                    return "";
                }
                else {
                    // Find the last colon in demangled, and use the rest
                    // as name
                    const unichar *demangled_u = (const unichar *) demangled;
                    const unichar *end = &(demangled_u[demangled.Length() - 1]);
                    while ((end > demangled_u) && (*end != ':')) {
                        end--;
                    }
                    if (*end == ':') {
                        end++;
                    }
                    return end;
                }
            }
            else {
                return name;
            }
        }

        // Returns the empty string if this class is anonymous, or has
        // a context whose name is the empty string
        String GetFullName()
        {
            if ((pContextM == NULL) || (pContextM->pContextM == NULL)) {
                return this->GetName();
            }
            else {
                return pContextM->GetFullName() + "::" + this->GetName();
            }
        }

        ref<class Context> GetContext()
        {
            return pContextM;
        }

        U32 GetFieldCount()
        {
            return vFieldsM.Count();
        }

        ref<class Field> GetField(U32 index)
        {
            return vFieldsM[index];
        }

        const char *GetAccessType()
        {
            String access = pElementM->GetAttributeValue("access");

            if (access == "private") {
                return "::Xrtti::AccessType_Private";
            }
            else if (access == "protected") {
                return "::Xrtti::AccessType_Protected";
            }
            else {
                return "::Xrtti::AccessType_Public";
            }
        }

    protected:

        Context(class Generator &generator, const String &id)
            : Registered(generator, id) { }

        void Initialize()
        {
            // Initialize the parent context
            if (pElementM->GetAttributeValue("name") != "::") {
                pContextM = generatorM.GetContext
                    (pElementM->GetAttributeValue("context"));
            }
            // Else the parent context is named "::", and so this is the
            // root namespace and doesn't have a parent context
        }

        void AddField(class Field *pField)
        {
            vFieldsM.Append(pField);
        }

    private:

        ref<class Context> pContextM;
        RefVector<Field> vFieldsM;
    };


    // -----------------------------------------------------------------------
    // This is a C++ namespace
    // -----------------------------------------------------------------------
    class Namespace : public Context
    {
    public:

        Namespace(class Generator &generator, const String &id)
            : Context(generator, id) { }
        // Potentially could add fields and methods of the namespace here; but
        // these are simply statically defined variables and methods outside
        // of any class definition, and not really the focus of the Xrtti
        // system; but this would have to be done in an Initialize method
        // anyway

        void Initialize()
        {
            this->Context::Initialize();
        }

        virtual Context::ContextType GetContextType()
        {
            return Context::ContextTypeNamespace;
        }

        virtual void Emit();
    };


    // -----------------------------------------------------------------------
    // This is an enumeration
    // -----------------------------------------------------------------------
    class Enumeration : public Registered
    {
    public:

        Enumeration(class Generator &generator, const String &id)
            : Registered(generator, id) { }
        
        void Initialize();

        void Emit();

        virtual RegisteredType GetRegisteredType()
        {
            return RegisteredTypeEnumeration;
        }
        
        const char *GetAccessType()
        {
            String access = pElementM->GetAttributeValue("access");

            if (access == "private") {
                return "::Xrtti::AccessType_Private";
            }
            else if (access == "protected") {
                return "::Xrtti::AccessType_Protected";
            }
            else {
                return "::Xrtti::AccessType_Public";
            }
        }

        ref<Context> GetContext()
        {
            return pContextM;
        }

        String GetName()
        {
            return pElementM->GetAttributeValue("name");
        }

        String GetFullName()
        {
            if ((pContextM == NULL) || (pContextM->GetContext() == NULL)) {
                return this->GetName();
            }
            else {
                return pContextM->GetFullName() + "::" + this->GetName();
            }
        }

        U32 GetEnumeratedValuesCount()
        {
            return vNamesM.Count();
        }

        String GetEnumeratedValuesName(U32 index)
        {
            return vNamesM[index];
        }

        S32 GetEnumeratedValuesValue(U32 index)
        {
            return vValuesM[index];
        }

    private:

        ref<Context> pContextM;
        Vector<String> vNamesM;
        Vector<S32> vValuesM;
    };


    // -----------------------------------------------------------------------
    // This is a C++ type, as used in method arguments, return types, and
    // class fields
    // -----------------------------------------------------------------------
    class Type : public Registered
    {
    public:

        // C++ types
        enum CppType
        {
            CppTypeVoid,
            CppTypeBool,
            CppTypeChar,
            CppTypeSignedChar,
            CppTypeUnsignedChar,
            CppTypeWChar,
            CppTypeShort,
            CppTypeSignedShort,
            CppTypeUnsignedShort,
            CppTypeInt,
            CppTypeSignedInt,
            CppTypeUnsignedInt,
            CppTypeLong,
            CppTypeSignedLong,
            CppTypeUnsignedLong,
            CppTypeLongLong,
            CppTypeSignedLongLong,
            CppTypeUnsignedLongLong,
            CppTypeFloat,
            CppTypeDouble,
            CppTypeLongDouble,
            CppTypeBitfield,
            CppTypeEnumeration,
            CppTypeFunction,
            CppTypeClass
        };

        // Attributes are used for the Type itself and each pointer "level"
        typedef struct Attributes
        {
            bool isConst;
            bool isVolatile;
        }
        Attributes;

        // MaxIndex is used to indicate the max index of an array
        typedef struct MaxIndex
        {
            bool isUnbounded; // If true, no max; if false, [max] is max
            U32 max;          // Only meaningful if [isUnbounded] is false
        }
        MaxIndex;

#if 0
        Type(class Generator &generator, const String &id, const String &bits)
            : Registered(generator, id), bitsM(bits) { }
#else
        Type(class Generator &generator, const String &id, const String &bits);
#endif

        void Initialize();
        
        void Emit();

        virtual RegisteredType GetRegisteredType()
        {
            return RegisteredTypeType;
        }

        CppType GetCppType()
        {
            return cppTypeM;
        }

        String GetCppFullType(bool includeConst);

        bool IsAssignable();
        String GetCastToValueFieldType();
        const char *GetCppXrttiName();
        String GetCppTypeName();
        String GetValueField();

        // If the CppType is CppTypeEnumeration, then this is the
        // Enumerated set of allowed values
        ref<Enumeration> GetCppTypeEnumeration()
        {
            return pCppTypeEnumerationM;
        }

        // If the CppType is CppTypeFunction, then this is the the function
        // type, as a Method
        ref<class Method> GetCppTypeMethod()
        {
            return pCppTypeMethodM;
        }

        // If the CppType is CppTypeClass, then this is the 
        // Class itself
        ref<class Class> GetCppTypeClass()
        {
            return pCppTypeClassM;
        }
        
        const Attributes &GetAttributes()
        {
            return attributesM;
        }

        bool IsReference()
        {
            return isReferenceM;
        }

        // Pointer depth; ie. "int ***pFoo" has pointer depth 3.
        U32 GetPointerDepth()
        {
            return vPointerAttributesM.Count();
        }

        // pointer attributes
        const Attributes &GetPointerAttributes(U32 index)
        {
            return vPointerAttributesM[index];
        }

        // Array Dimensions count
        U32 GetArrayDimensionsCount()
        {
            return vArrayDimensionMaxIndexesM.Count();
        }

        // Array Dimensions Max Index
        const MaxIndex &GetArrayDimensionMaxIndex(U32 index)
        {
            return vArrayDimensionMaxIndexesM[index];
        }

    private:

        constref<Parser::Element> GetCppTypeElement(const Parser::Element *p);
    
        String bitsM;
        CppType cppTypeM;
        ref<Enumeration> pCppTypeEnumerationM;
        ref<class Method> pCppTypeMethodM;
        ref<class Generator::Class> pCppTypeClassM;
        Attributes attributesM;
        bool isReferenceM;
        Vector<Attributes> vPointerAttributesM;
        Vector<MaxIndex> vArrayDimensionMaxIndexesM;
    };

    // -----------------------------------------------------------------------
    // This is a C++ class field (member variable)
    // -----------------------------------------------------------------------
    class Field : public Base
    {
    public:

        Field(class Generator &generator, const String &id,
              Class *pContainingClass, bool isStatic);

        void Emit();

        String GetName()
        {
            return pElementM->GetAttributeValue("name");
        }

        U32 GetNumber()
        {
            return numberM;
        }

        ref<Type> GetType()
        {
            return pTypeM;
        }

        bool IsStatic()
        {
            return isStaticM;
        }

        const char *GetAccessType()
        {
            String access = pElementM->GetAttributeValue("access");

            if (access == "private") {
                return "::Xrtti::AccessType_Private";
            }
            else if (access == "protected") {
                return "::Xrtti::AccessType_Protected";
            }
            else {
                return "::Xrtti::AccessType_Public";
            }
        }

    private:

        String GetGetMethodName()
        {
            return String::Format("XrttiPrivate_Field_%lu_get", numberM);
        }
        
        String GetSetMethodName()
        {
            return String::Format("XrttiPrivate_Field_%lu_set", numberM);
        }

        void EmitGet();
        void EmitSet();

        ref<Type> pTypeM;
        U32 numberM;
        Class *pContainingClassM;
        bool isStaticM;
        String offsetofStringM;

        static U32 fieldNumberG;
    };

    // -----------------------------------------------------------------------
    // This is a method argument
    // -----------------------------------------------------------------------
    class Argument : public RefCounted
    {
    public:

        Argument(class Generator &generator, const String &name,
                 const String &typeId, bool hasDefault, 
                 const String &defaultValue)
            : nameM(name), defaultM(defaultValue), hasDefaultM(hasDefault)
        {
            pTypeM = generator.GetType(typeId, "");
        }

        String GetName()
        {
            return nameM;
        }

        ref<Type> GetType()
        {
            return pTypeM;
        }

        bool HasDefault()
        {
            return hasDefaultM;
        }

        String GetDefault()
        {
            return defaultM;
        }

    private:

        String nameM, defaultM;
        ref<Type> pTypeM;
        bool hasDefaultM;
    };

    // -----------------------------------------------------------------------
    // This is a method, constructor, or function
    // -----------------------------------------------------------------------
    class Method : public Base
    {
    public:

        enum MethodType
        {
            MethodTypeConstructor,
            MethodTypeDestructor,
            MethodTypeMethod,
            MethodTypeOperatorMethod,
            MethodTypeFunctionSignature
        };
        
        Method(class Generator &generator, const String &id, MethodType type);

        void Emit(Class *pContainingClass);
        
        U32 GetNumber()
        {
            return numberM;
        }

        String GetName()
        {
            return pElementM->GetAttributeValue("name");
        }

        bool IsStatic()
        {
            return (pElementM->GetAttributeValue("static") == "1");
        }

        const char *GetAccessType()
        {
            String access = pElementM->GetAttributeValue("access");

            if (access == "private") {
                return "::Xrtti::AccessType_Private";
            }
            else if (access == "protected") {
                return "::Xrtti::AccessType_Protected";
            }
            else {
                return "::Xrtti::AccessType_Public";
            }
        }

        bool IsConst()
        {
            return (pElementM->GetAttributeValue("const") == "1");
        }

        bool IsVirtual()
        {
            return (pElementM->GetAttributeValue("virtual") == "1");
        }
        
        bool IsPureVirtual()
        {
            return (pElementM->GetAttributeValue("pure_virtual") == "1");
        }

        U32 GetArgumentCount()
        {
            return vArgumentsM.Count();
        }

        ref<Argument> GetArgument(U32 index)
        {
            return vArgumentsM[index];
        }

        bool HasEllipsis()
        {
            return hasEllipsisM;
        }

        ref<Type> GetReturnType()
        {
            return pReturnTypeM;
        }

        void EmitSignature();
        String GetSignature();

    private:

        void EmitInvoke(Class *pContainingContext);

        MethodType methodTypeM;
        RefVector<Argument> vArgumentsM;
        bool hasEllipsisM;
        ref<Type> pReturnTypeM;
        RefVector<Type> vThrowsM;
        U32 numberM;

        static U32 methodNumberG;
    };

    // -----------------------------------------------------------------------
    // Class is a class, struct, or union.  Each Class has a FullName, which
    // is the name of the actual type represented by the class (and which
    // may not be a valid C++ name, if the class is declared within an
    // anonymous class).
    // -----------------------------------------------------------------------
    class Class : public Context
    {
    public:

        Class(class Generator &generator, const String &id)
            : Context(generator, id) { }
        
        void Initialize();

        virtual Context::ContextType GetContextType()
        {
            return Context::ContextTypeClass;
        }

        String GetTypedefName()
        {
            return String::Format("type_%lu", this->GetNumber());
        }

        void EmitTypedef();

        virtual void Emit();

        // An anonymous class is one without a name; there are two cases of
        // anonymous classes:
        //
        // 1. class foo { class { int a; } anonymous; }
        // 2. class foo { class { int a; }; };
        //
        // In the first case, the class has a "defining field", which is
        // the single field of the containing context whose type is this
        // class.  In the second case, the class has no defining field, and
        // in fact every field in the class can and should be treated just
        // like it is a direct member of the containing class, for all
        // intents and purposes.
        bool IsAnonymous()
        {
            String name = pElementM->GetAttributeValue("name");

            if ((name == "") || name.StartsWith(".")) {
                return true;
            }

            ref<Context> pContext = this->GetContext();
            
            if ((pContext == NULL) || 
                (pContext->GetContextType() != Context::ContextTypeClass)) {
                return false;
            }
            
            ref<Class> pClass = (Class *) (Context *) pContext;
            
            return pClass->IsAnonymous();
        }
        
        const char *GetContextTypeName()
        {
            return (this->IsIncomplete() ? 
                    "::Xrtti::Context::Type_Incomplete" :
                    this->IsStruct() ? "::Xrtti::Context::Type_Struct" :
                    this->IsUnion() ? "::Xrtti::Context::Type_Union" :
                    "::Xrtti::Context::Type_Class");
        }

        bool IsIncomplete()
        {
            return (pElementM->GetAttributeValue("incomplete") == "1");
        }

        bool IsAbstract();

        bool IsStruct()
        {
            return (pElementM->GetElementName() == "Struct");
        }

        bool IsUnion()
        {
            return (pElementM->GetElementName() == "Union");
        }

        U32 GetSuperCount()
        {
            return vSupersM.Count();
        }

        ref<Class> GetSuper(U32 index)
        {
            return vSupersM[index];
        }

        // constructors enumeration
        U32 GetConstructorCount()
        {
            return vConstructorsM.Count();
        }

        ref<Method> GetConstructor(U32 index)
        {
            return vConstructorsM[index];
        }

        bool HasDestructor()
        {
            return (pDestructorM == NULL) ? false : true;
        }
        
        ref<Method> GetDestructor()
        {
            return pDestructorM;
        }

        U32 GetMethodCount()
        {
            return vMethodsM.Count();
        }

        ref<Method> GetMethod(U32 index)
        {
            return vMethodsM[index];
        }

    private:

        bool CanBeCreatedByDefaultConstructor();
        bool CanBeDestroyedByDefaultDestructor();
        void EmitFields();
        void EmitConstructors();
        void EmitDestructor();
        void EmitMethods();
        void EmitVector(U32 count, const char *name,
                        bool emitCount, bool emitComma);

        RefVector<Class> vSupersM;
        RefVector<Method> vConstructorsM;
        ref<Method> pDestructorM;
        RefVector<Method> vMethodsM;
    };

    // -----------------------------------------------------------------------
    // Methods
    // -----------------------------------------------------------------------
    ref<Parser::Element> LookupParserElement(const String &id)
    {
        return parserM.LookupElement(id);
    }

    OutputStream &GetOutputStream()
    {
        return outM;
    }

    void Error(const String &format, ...);
    bool GetRtti() { return configM.GetRtti(); }
    
private:

    ref<Context> GetContext(const String &id);
    ref<Enumeration> GetEnumeration(const String &id);
    ref<Type> GetType(const String &id, const String &bits);
    void EmitHeader();
    void EmitTypedefs();
    void EmitGenerated();
    void EmitSpecialArray(bool trueArray);
    void EmitSpecialArrayReference(bool trueArray, const char *name);
    static void split(const unichar *pString, RefVector<RefString> &v);
    String GetFullNameOf(const String &elementId);

    // -----------------------------------------------------------------------
    // Members
    // -----------------------------------------------------------------------
    const Configuration &configM;
    const Parser &parserM;
    OutputStream &outM;
    bool isTrueArrayWrittenM, isFalseArrayWrittenM;
    // Hashtables for fast lookup by id
    RefHashtable<String, Context> htContextsM;
    RefHashtable<String, Enumeration> htEnumerationsM;
    RefHashtable<String, Type> htTypesM;
};

#endif // GENERATOR_H
